from collections import OrderedDict
import os
import unicodedata


def load_vocab(vocab_file):
    if not os.path.isfile(vocab_file):
        raise ValueError(
            f"Can't find a vocabulary file at path '{vocab_file}'.")

    vocab = OrderedDict()
    with open(vocab_file, "r", encoding="utf-8") as reader:
        tokens = reader.readlines()
    for index, token in enumerate(tokens):
        token = token.rstrip("\n")
        vocab[token] = index
    return vocab


def whitespace_tokenize(text):
    text = text.strip()
    if not text:
        return []
    tokens = text.split()
    return tokens


def _is_whitespace(char):
    if char == " " or char == "\t" or char == "\n" or char == "\r":
        return True
    cat = unicodedata.category(char)
    if cat == "Zs":
        return True
    return False


def _is_control(char):
    if char == "\t" or char == "\n" or char == "\r":
        return False
    cat = unicodedata.category(char)
    if cat.startswith("C"):
        return True
    return False


def _is_punctuation(char):
    cp = ord(char)
    if (cp >= 33 and cp <= 47) or (cp >= 58 and cp <= 64) or (
            cp >= 91 and cp <= 96) or (cp >= 123 and cp <= 126):
        return True
    cat = unicodedata.category(char)
    if cat.startswith("P"):
        return True
    return False


from .utils import whitespace_tokenize, _is_whitespace, _is_punctuation, _is_control
import unicodedata


class BasicTokenizer:
    """Constructs a BasicTokenizer that will run basic tokenization (punctuation splitting, lower casing, etc.)."""

    def tokenize(self, text):
        text = self._clean_text(text)
        text = self._pad_chinese_chars(text)
        orig_tokens = whitespace_tokenize(text)
        split_tokens = []
        for token in orig_tokens:
            token = token.lower()
            token = self._run_strip_accents(token)
            split_tokens.extend(self._run_split_on_punc(token))

        output_tokens = whitespace_tokenize(" ".join(split_tokens))
        return output_tokens

    def _clean_text(self, text):
        output = []
        for char in text:
            cp = ord(char)
            if cp == 0 or cp == 0xFFFD or _is_control(char):
                continue
            if _is_whitespace(char):
                output.append(" ")
            else:
                output.append(char)
        return "".join(output)

    def _run_strip_accents(self, text):
        text = unicodedata.normalize("NFD", text)
        output = []
        for char in text:
            cat = unicodedata.category(char)
            if cat == "Mn":
                continue
            output.append(char)
        return "".join(output)

    def _run_split_on_punc(self, text):
        chars = list(text)
        i = 0
        start_new_word = True
        output = []
        while i < len(chars):
            char = chars[i]
            if _is_punctuation(char):
                if char == '[' and i + 5 < len(chars) and "".join(chars[i:i+6]).upper() == "[MASK]":
                    output.append([x.upper() for x in chars[i:i+6]])
                    i += 5
                else:
                    output.append([char])
                start_new_word = True
            else:
                if start_new_word:
                    output.append([])
                start_new_word = False
                output[-1].append(char)
            i += 1

        return ["".join(x) for x in output]

    def _pad_chinese_chars(self, text):
        output = []
        for char in text:
            cp = ord(char)
            if self._is_chinese_char(cp):
                output.append(" ")
                output.append(char)
                output.append(" ")
            else:
                output.append(char)
        return "".join(output)

    def _is_chinese_char(self, cp):
        if ((cp >= 0x4E00 and cp <= 0x9FFF)  #
                or (cp >= 0x3400 and cp <= 0x4DBF)  #
                or (cp >= 0x20000 and cp <= 0x2A6DF)  #
                or (cp >= 0x2A700 and cp <= 0x2B73F)  #
                or (cp >= 0x2B740 and cp <= 0x2B81F)  #
                or (cp >= 0x2B820 and cp <= 0x2CEAF)  #
                or (cp >= 0xF900 and cp <= 0xFAFF)  #
                or (cp >= 0x2F800 and cp <= 0x2FA1F)  #
            ):
            return True

        return False



from collections import OrderedDict
from pathlib import Path
import os

from .BasicTokenizer import BasicTokenizer
from .Trie import Trie
from .utils import load_vocab


class WordPieceTokenizer:

    def __init__(self):
        self.unk_token = "[UNK]"
        self.sep_token = "[SEP]"
        self.cls_token = "[CLS]"

        self.vocab = load_vocab(os.path.join(str(Path(__file__).resolve().parent), "vocab.txt"))
        self._initialise_tokens_trie()

        self.ids_to_tokens = OrderedDict([(ids, tok)
                                          for tok, ids in self.vocab.items()])
        self.basic_tokenizer = BasicTokenizer()

    def _initialise_tokens_trie(self):
        self.tokens_trie = Trie(self._convert_token_to_id(self.unk_token))
        for tok, tok_id in self.vocab.items():
            self.tokens_trie.add(tok, tok_id)

    def tokenize(self, text):
        cls_token_id = self._convert_token_to_id(self.cls_token)
        sep_token_id = self._convert_token_to_id(self.sep_token)
        tokenized_text = [cls_token_id, *self._tokenize(text), sep_token_id]
        return tokenized_text

    def _tokenize(self, text):
        split_tokens = []
        for token in self.basic_tokenizer.tokenize(text):
            split_tokens += self._wordpiece_tokenize(token)
        return split_tokens

    def _wordpiece_tokenize(self, text):
        token_ids = []

        while text != "##":
            token_id, text = self.tokens_trie.getLongestMatchToken(text)
            text = "##" + text
            token_ids = [*token_ids, token_id]

        return token_ids

    def _convert_token_to_id(self, token):
        return self.vocab.get(token, self.vocab.get(self.unk_token))

    def _convert_id_to_token(self, index):
        return self.ids_to_tokens.get(index, self.unk_token)

    def convert_tokens_to_string(self, tokens):
        out_string = " ".join(tokens).replace(" ##", "").strip()
        return out_string
    
    def convert_ids_to_tokens(self, ids):
        return [self._convert_id_to_token(tok_id) for tok_id in ids]



from typing import Union


class Trie:

    def __init__(self, unk_token_id):
        self.data = {}
        self.unk_token_id = unk_token_id

    def add(self, token: str, token_id: str):
        if not token:
            return
        ref = self.data
        for char in token:
            ref[char] = char in ref and ref[char] or {}
            ref = ref[char]
        ref[""] = token_id

    def getLongestMatchToken(self, text: str) -> Union[int, str]:
        """Returns token of the longest match and the remaining unmatched text."""
        curr_index = longest_match_end_index = 0
        longest_match_token = -1
        ref = self.data

        while curr_index < len(text) and text[curr_index] in ref:
            ref = ref[text[curr_index]]
            if ("" in ref and (len(text) == 1 or text[curr_index] != "#")):
                longest_match_end_index = curr_index
                longest_match_token = ref[""]
            curr_index += 1
        if longest_match_token == -1:
            return self.unk_token_id, text[1:]
        else:
            return longest_match_token, text[longest_match_end_index + 1:]
